from abc import abstractmethod
import copy
import numpy as np
import time
import matplotlib.pyplot as plt

from sklearn.model_selection import train_test_split
from sklearn import datasets
from sklearn.preprocessing import LabelEncoder

from wrapper.nature_inspired._utilities_test import Solution, Data, compute_accuracy, compute_fitness, initialize, sort_agents, display, call_counter


class Algorithm():
    def __init__(self,
                 num_agents,
                 max_iter,
                 train_data,
                 train_label,
                 test_data=None,
                 test_label=None,
                 val_size=30,
                 seed=0,
                 save_conv_graph=True,
                 max_evals=np.float("inf"),
                 algo_name=None,
                 agent_name=None):

        # essential user-defined variables
        self.num_agents = num_agents
        self.max_iter = max_iter
        self.training_data = None
        self.train_data = train_data
        self.train_label = train_label

        # user-defined or default variables
        self.test_data = test_data
        self.test_label = test_label
        self.val_size = val_size
        self.weight_acc = None
        self.seed = seed
        self.save_conv_graph = save_conv_graph

        # algorithm internal variables
        self.population = None
        self.num_features = None
        self.fitness = None
        self.accuracy = None
        self.Leader_agent = None
        self.Leader_fitness = float("-inf")
        self.Leader_accuracy = float("-inf")
        self.history = []
        self.cur_iter = 0
        self.max_evals = max_evals
        self.start_time = None
        self.end_time = None
        self.exec_time = None
        self.solution = None
        self.algo_name = algo_name
        self.agent_name = agent_name


    @abstractmethod
    def user_input(self):
        pass


    @abstractmethod
    def next(self):
        pass

    
    def int_encoding(self, labels):
        # converts the labels to one-hot-encoded vectors
        labels_str = np.array([str(i) for i in labels])

        # integer encode
        label_encoder = LabelEncoder()
        integer_encoded = label_encoder.fit_transform(labels_str)

        # # binary encode
        # onehot_encoder = OneHotEncoder(sparse=False)
        # integer_encoded = integer_encoded.reshape(len(integer_encoded), 1)
        # onehot_encoded = onehot_encoder.fit_transform(integer_encoded)
        
        return integer_encoded


    def initialize(self):
        # set the objective function
        self.val_size = float(input('Percentage of data for valdiation [0-100]: ') or 30)/100
        self.weight_acc = float(input('Weight for the classification accuracy [0-1]: ') or 0.9)
        self.obj_function = call_counter(compute_fitness(self.weight_acc))

        # start timer
        self.start_time = time.time()
        np.random.seed(self.seed)

        # data preparation
        self.training_data = Data()
        self.train_data, self.train_label = np.array(self.train_data), np.array(self.train_label)
        self.train_label = self.int_encoding(self.train_label)
        self.training_data.train_X, self.training_data.val_X, self.training_data.train_Y, self.training_data.val_Y = train_test_split(self.train_data, self.train_label, stratify=self.train_label, test_size=self.val_size)

        # create initial population
        self.num_features = self.train_data.shape[1]
        self.population = initialize(num_agents=self.num_agents, num_features=self.num_features)
        self.fitness = self.obj_function(self.population, self.training_data)
        self.population, self.fitness = sort_agents(agents=self.population, fitness=self.fitness)
        self.accuracy = compute_accuracy(agents=self.population, data=self.training_data)
        self.Leader_agent, self.Leader_fitness = self.population[0], self.fitness[0]

    def check_end(self):
        # checks if the algorithm has met the end criterion
        return (self.cur_iter >= self.max_iter) or (self.obj_function.cur_evals >= self.max_evals)


    def save_details(self):
        # save some details of every generation
        cur_obj = {
            'population': self.population,
            'fitness': self.fitness,
            'accurcay': self.accuracy,
        }
        self.history.append(cur_obj)


    def display(self):
        # display the current generation details
        display(agents=self.population, fitness=self.fitness, agent_name=self.agent_name)


    def plot(self):
        # plot the convergence graph
        fig = plt.figure(figsize=(10, 8))
        avg_fitness = []
        for cur in self.history:
            avg_fitness.append(np.mean(cur['fitness']))

        plt.plot(np.arange(len(avg_fitness)), avg_fitness)
        plt.xlabel('Number of Generations')
        plt.ylabel('Average Fitness')
        plt.title('Convergence Curve')

        plt.show()

        return fig

    
    def post_processing(self):
        # post processing steps
        self.fitness = self.obj_function(self.population, self.training_data)
        self.population, self.fitness = sort_agents(agents=self.population, fitness=self.fitness)
        self.accuracy = compute_accuracy(agents=self.population, data=self.training_data)
        
        if(self.fitness[0] > self.Leader_fitness):
            self.Leader_fitness = self.fitness[0]
            self.Leader_agent = self.population[0, :]
            self.Leader_accuracy = self.accuracy[0]


    def save_solution(self):
        # create a solution object
        self.solution = Solution()
        self.solution.num_agents = self.num_agents
        self.solution.max_iter = self.max_iter
        self.solution.num_features = self.train_data.shape[1]
        self.solution.obj_function = self.obj_function

        # update attributes of solution
        self.solution.best_agent = self.Leader_agent
        self.solution.best_fitness = self.Leader_fitness
        self.solution.best_accuracy = self.Leader_accuracy
        self.solution.final_population = self.population
        self.solution.final_fitness = self.fitness
        self.solution.final_accuracy = self.accuracy
        self.solution.execution_time = self.exec_time


    def run(self):
        # the main algorithm run
        print('\n************    Please enter the values of the following paramters or press newline for using default values    ************\n')
        self.user_input()   # take the user inputs
        self.initialize()   # initialize the algorithm
        print('\n*****************************************************    Thank You    ******************************************************\n')

        while(not self.check_end()):    # while the end criterion is not met
            self.next()                     # do one step of the algorithm
            self.post_processing()          # do the post processing steps
            self.display()                  # display the details of 1 iteration
            self.save_details()             # save the details

        self.end_time = time.time()     
        self.exec_time = self.end_time - self.start_time

        if self.test_data:          # if there is a test data, test the final solution on that 
            self.test_label = self.int_encoding(self.test_label)
            temp_data = Data()
            temp_data.train_X = self.train_data
            temp_data.train_Y = self.train_label
            temp_data.val_X = self.test_data
            temp_data.val_Y = self.test_label

            self.Leader_fitness = compute_fitness(self.Leader_agent, temp_data)
            self.Leader_accuracy = compute_accuracy(self.Leader_agent, temp_data)

        self.save_solution()
        fig = self.plot()

        if(self.save_conv_graph):
            fig.savefig('convergence_curve_' + self.algo_name + '.jpg')

        print('\n------------- Leader Agent ---------------')
        print('Fitness: {}'.format(self.Leader_fitness))
        print('Number of Features: {}'.format(int(np.sum(self.Leader_agent))))
        print('----------------------------------------\n')

        return self.solution

 








"""

Programmer: Ritam Guha
Date of Development: 6/10/2020

"""
import numpy as np
from wrapper.nature_inspired.algorithm import Algorithm
from sklearn import datasets

class GA(Algorithm):

    # Genetic Algorithm
    ############################### Parameters ####################################
    #                                                                             #
    #   num_agents: number of chromosomes                                         #
    #   max_iter: maximum number of generations                                   #
    #   train_data: training samples of data                                      #
    #   train_label: class labels for the training samples                        #                
    #   obj_function: the function to maximize while doing feature selection      #
    #   prob_cross: probability of crossover                                      #
    #   prob_mut: probability of mutation                                         #
    #   save_conv_graph: boolean value for saving convergence graph               #
    #                                                                             #
    ###############################################################################

    def __init__(self,   
                num_agents, 
                max_iter, 
                train_data, 
                train_label, 
                save_conv_graph=False, 
                seed=0):

        super().__init__( num_agents=num_agents,
                        max_iter=max_iter,
                        train_data=train_data,
                        train_label=train_label,
                        save_conv_graph=save_conv_graph,
                        seed=seed )

        self.algo_name = 'GA'
        self.agent_name = 'Chromosome'
        self.algo_params = {}


    def user_input(self):
        # accept the parameters as user inputs
        self.algo_params['prob_cross'] = float(input('Probability of crossover [0-1]: ') or 0.7)
        self.algo_params['prob_mut'] = float(input('Probability of mutation [0-1]: ') or 0.3)
        self.algo_params['cross_limit'] = float(input('Max crossover in every Generation [5-10]: ') or 5)
        

    def crossover(self, parent_1, parent_2):
        # perform crossover with crossover probability prob_cross
        num_features = parent_1.shape[0]
        child_1 = parent_1.copy()
        child_2 = parent_2.copy()

        for i in range(num_features):
            if(np.random.rand() < self.algo_params['prob_cross']):
                child_1[i] = parent_2[i]
                child_2[i] = parent_1[i]

        return child_1, child_2


    def mutation(self, chromosome):
        # perform mutation with mutation probability prob_mut
        num_features = chromosome.shape[0]
        mut_chromosome = chromosome.copy()

        for i in range(num_features):
            if(np.random.rand() < self.algo_params['prob_mut']):
                mut_chromosome[i] = 1-mut_chromosome[i]
        
        return mut_chromosome


    def roulette_wheel(self, fitness):
        # perform roulette wheel selection
        maximum = sum([f for f in fitness])
        selection_probs = [f/maximum for f in fitness]
        return np.random.choice(len(fitness), p=selection_probs)


    def cross_mut(self):
        # perform crossover, mutation and replacement
        count = 0
        print('Crossover-Mutation phase starting....')

        while(count < self.algo_params['cross_limit']):
            print('\nCrossover no. {}'.format(count+1))
            id_1 = self.roulette_wheel(self.fitness)
            id_2 = self.roulette_wheel(self.fitness)

            if(id_1 != id_2):
                child_1, child_2 = self.crossover(self.population[id_1, :], self.population[id_2, :])
                child_1 = self.mutation(child_1)
                child_2 = self.mutation(child_2)

                child = np.array([child_1, child_2])
                child_fitness = self.obj_function(child, self.training_data)
                child, child_fitness = sort_agents(child, child_fitness)

                for i in range(2):
                    for j in reversed(range(self.num_agents)):
                        if(child_fitness[i] > self.fitness[j]):
                            print('child {} replaced with chromosome having id {}'.format(i+1, j+1))
                            self.population[j, :] = child[i]
                            self.fitness[j] = child_fitness[i]
                            break
                
                count = count+1

            else:
                print('Crossover failed....')
                print('Restarting crossover....\n')

    
    def next(self):
        print('\n================================================================================')
        print('                          Iteration - {}'.format(self.cur_iter+1))
        print('================================================================================\n')

        # perform crossover, mutation and replacement
        self.cross_mut()

        self.cur_iter += 1



############# for testing purpose ################
if __name__ == '__main__':
    data = datasets.load_digits()
    algo = GA(num_agents=20, max_iter=100, train_data=data.data, train_label=data.target)
    solution = algo.run()
############# for testing purpose ################



import pandas as pd
import numpy as np
import glob, os
from tabulate import tabulate
import pkg_resources

class Data():
    def __init__(self):
        self.data = None
        self.target = None

def get_dataset(dataset_name):
    dir_name = os.path.dirname(os.path.abspath(__file__))
    data_paths = pkg_resources.resource_listdir('Py_FS.datasets','database')
    # print("Package Paths:", data_paths)
    list_datasets = [os.path.splitext(os.path.basename(filename))[0] for filename in data_paths]

    # print(dir_name)

    if dataset_name not in list_datasets:
        print(f"[!Error] Py_FS currently does not have {dataset_name} in its database....")
        display = input("Enter 1 to see the available datasets: ") or 0
        if display:
            display_datasets(list_datasets)

    else:
        data = Data()
        stream = pkg_resources.resource_stream(__name__, 'database/' + dataset_name + '.csv')
        df = pd.read_csv(stream, header=None)
        data.data = np.array(df.iloc[:, 0:-1])
        data.target = np.array(df.iloc[:, -1])

        print(data.data.shape)
        print(data.target.shape)


def display_datasets(list_datasets):
    print("\n=========== Available Datasets ===========\n")
    table_list = []

    for i, dataset in enumerate(list_datasets):
        table_list.append([i+1, list_datasets[i]])

    print(tabulate(table_list, headers=["Index", "Dataset"]))

if __name__ == '__main__':
    get_dataset('Vow')



"""

Programmer: Ritam Guha
Date of Development: 8/10/2020
This code has been developed according to the procedures mentioned in the following research article:
"Mafarja, M., & Mirjalili, S. (2018). Whale optimization approaches for wrapper feature selection. 
Applied Soft Computing, 62, 441-453."

"""

import numpy as np
from wrapper.nature_inspired.algorithm import Algorithm
from wrapper.nature_inspired._transfer_functions import get_trans_function
from sklearn import datasets

class WOA(Algorithm):

    # Whale Optimization Algorithm
    ############################### Parameters ####################################
    #                                                                             #
    #   num_agents: number of whales                                              #
    #   max_iter: maximum number of generations                                   #
    #   train_data: training samples of data                                      #
    #   train_label: class labels for the training samples                        #                
    #   obj_function: the function to maximize while doing feature selection      #
    #   trans_function_shape: shape of the transfer function used                 #
    #   save_conv_graph: boolean value for saving convergence graph               #
    #                                                                             #
    ###############################################################################
    
    def __init__( self,   
                num_agents, 
                max_iter, 
                train_data, 
                train_label, 
                save_conv_graph=False, 
                seed=0):

        super().__init__( num_agents=num_agents,
                        max_iter=max_iter,
                        train_data=train_data,
                        train_label=train_label,
                        save_conv_graph=save_conv_graph,
                        seed=seed )

        self.algo_name = 'WOA'
        self.agent_name = 'Whale'
        self.trans_function = None
        self.algo_params = {}

    
    def user_input(self):
        # accept the parameters as user inputs
        self.algo_params['trans_function'] = input('Shape of Transfer Function [s/v/u]: ') or 's'
        self.trans_function = get_trans_function(self.algo_params['trans_function'])

    
    def forage(self):
        a = 2 - self.cur_iter * (2/self.max_iter)  # a decreases linearly fron 2 to 0
        # update the position of each whale
        for i in range(self.num_agents):
            # update the parameters
            r = np.random.rand() # r is a random number in [0, 1]
            A = (2 * a * r) - a  # Eq. (3)
            C = 2 * r  # Eq. (4)
            l = np.random.uniform(-1, 1)   # l is a random number in [-1, 1]
            p = np.random.random()  # p is a random number in [0, 1]
            b = 1  # defines shape of the spiral               
            
            if p<0.5:
                # Shrinking Encircling mechanism
                if abs(A)>=1:
                    rand_agent_index = np.random.randint(0, self.num_agents)
                    rand_agent = self.population[rand_agent_index, :]
                    mod_dist_rand_agent = abs(C * rand_agent - self.population[i, :]) 
                    self.population[i, :] = rand_agent - (A * mod_dist_rand_agent)   # Eq. (9)
                    
                else:
                    mod_dist_Leader = abs(C * self.Leader_agent - self.population[i, :]) 
                    self.population[i, :] = self.Leader_agent - (A * mod_dist_Leader)  # Eq. (2)
                
            else:
                # Spiral-Shaped Attack mechanism
                dist_Leader = abs(self.Leader_agent - self.population[i, :])
                self.population[i, :] = dist_Leader * np.exp(b * l) * np.cos(l * 2 * np.pi) + self.Leader_agent

            # Apply transformation function on the updated whale
            for j in range(self.num_features):
                trans_value = self.trans_function(self.population[i, j])
                if (np.random.rand() < trans_value): 
                    self.population[i, j] = 1
                else:
                    self.population[i, j] = 0


    # main loop
    def next(self):
        print('\n================================================================================')
        print('                          Iteration - {}'.format(self.cur_iter+1))
        print('================================================================================\n')

        # perform bubble-net foraging actions
        self.forage()

        self.cur_iter += 1

############# for testing purpose ################

if __name__ == '__main__':
    data = datasets.load_digits()
    algo = WOA(num_agents=20, max_iter=100, train_data=data.data, train_label=data.target, save_conv_graph=True)
    algo.run()
############# for testing purpose ################



"""

Programmer: Ali Hussain Khan
Date of Development: 15/10/2020
This code has been developed according to the procedures mentioned in the following research article:
"Khanesar, M. A., Teshnehlab, M., & Shoorehdeli, M. A. (2007, June). 
A novel binary particle swarm optimization. 
In 2007 Mediterranean conference on control & automation (pp. 1-6). IEEE."

"""
import numpy as np
import time
import matplotlib.pyplot as plt

from sklearn.model_selection import train_test_split
from sklearn import datasets

from wrapper.nature_inspired._utilities import Solution, Data, initialize, sort_agents, display, compute_fitness, Conv_plot
from wrapper.nature_inspired._transfer_functions import get_trans_function


def PSO(num_agents, max_iter, train_data, train_label, obj_function=compute_fitness, trans_func_shape='s', save_conv_graph=False):
    
    # Particle Swarm Optimizer
    ############################### Parameters ####################################
    #                                                                             #
    #   num_agents: number of particles                                           #
    #   max_iter: maximum number of generations                                   #
    #   train_data: training samples of data                                      #
    #   train_label: class labels for the training samples                        #                
    #   obj_function: the function to maximize while doing feature selection      #
    #   trans_function_shape: shape of the transfer function used                 #
    #   save_conv_graph: boolean value for saving convergence graph               #
    #                                                                             #
    ###############################################################################
    
    short_name = 'PSO'
    agent_name = 'Particle'
    train_data, train_label = np.array(train_data), np.array(train_label)
    num_features = train_data.shape[1]
    trans_function = get_trans_function(trans_func_shape)
    
    # setting up the objectives
    weight_acc = None
    if(obj_function==compute_fitness):
        weight_acc = float(input('Weight for the classification accuracy [0-1]: '))
    obj = (obj_function, weight_acc)
    compute_accuracy = (compute_fitness, 1) # compute_accuracy is just compute_fitness with accuracy weight as 1

    # initialize particles and Leader (the agent with the max fitness)
    particles = initialize(num_agents, num_features)
    fitness = np.zeros(num_agents)
    accuracy = np.zeros(num_agents)
    Leader_agent = np.zeros((1, num_features))
    Leader_fitness = float("-inf")
    Leader_accuracy = float("-inf")

    # initialize convergence curves
    convergence_curve = {}
    convergence_curve['fitness'] = np.zeros(max_iter)

    # initialize data class
    data = Data()
    val_size = float(input('Enter the percentage of data wanted for valdiation [0, 100]: '))/100
    data.train_X, data.val_X, data.train_Y, data.val_Y = train_test_split(train_data, train_label, stratify=train_label, test_size=val_size)

    # create a solution object
    solution = Solution()
    solution.num_agents = num_agents
    solution.max_iter = max_iter
    solution.num_features = num_features
    solution.obj_function = obj_function

    # rank initial particles
    particles, fitness = sort_agents(particles, obj, data)

    # start timer
    start_time = time.time()

    # initialize global and local best particles
    globalBestParticle = [0 for i in range(num_features)]
    globalBestFitness = float("-inf")
    localBestParticle = [ [ 0 for i in range(num_features) ] for j in range(num_agents) ] 
    localBestFitness = [float("-inf") for i in range(num_agents) ]
    weight = 1.0 
    velocity = [ [ 0.0 for i in range(num_features) ] for j in range(num_agents) ]
    
    for iter_no in range(max_iter):
        print('\n================================================================================')
        print('                          Iteration - {}'.format(iter_no+1))
        print('================================================================================\n')
        
        # update weight
        weight = 1.0 - (iter_no / max_iter)
        
        # update the velocity
        for i in range(num_agents):
            for j in range(num_features):
                velocity[i][j] = (weight*velocity[i][j])
                r1, r2 = np.random.random(2)
                velocity[i][j] = velocity[i][j] + (r1 * (localBestParticle[i][j] - particles[i][j]))
                velocity[i][j] = velocity[i][j] + (r2 * (globalBestParticle[j] - particles[i][j]))
       
        # updating position of particles
        for i in range(num_agents):
            for j in range(num_features):
                trans_value = trans_function(velocity[i][j])
                if (np.random.random() < trans_value): 
                    particles[i][j] = 1
                else:
                    particles[i][j] = 0
                 
        # updating fitness of particles
        particles, fitness = sort_agents(particles, obj, data)
        display(particles, fitness, agent_name)
        
        
        # updating the global best and local best particles
        for i in range(num_agents):
            if fitness[i]>localBestFitness[i]:
                localBestFitness[i]=fitness[i]
                localBestParticle[i]=particles[i][:]

            if fitness[i]>globalBestFitness:
                globalBestFitness=fitness[i]
                globalBestParticle=particles[i][:]

        # update Leader (best agent)
        if globalBestFitness > Leader_fitness:
            Leader_agent = globalBestParticle.copy()
            Leader_fitness = globalBestFitness.copy()

        convergence_curve['fitness'][iter_no] = np.mean(fitness)

    # compute final accuracy
    Leader_agent, Leader_accuracy = sort_agents(Leader_agent, compute_accuracy, data)
    particles, accuracy = sort_agents(particles, compute_accuracy, data)

    print('\n================================================================================')
    print('                                    Final Result                                  ')
    print('================================================================================\n')
    print('Leader ' + agent_name + ' Dimension : {}'.format(int(np.sum(Leader_agent))))
    print('Leader ' + agent_name + ' Fitness : {}'.format(Leader_fitness))
    print('Leader ' + agent_name + ' Classification Accuracy : {}'.format(Leader_accuracy))
    print('\n================================================================================\n')

    # stop timer
    end_time = time.time()
    exec_time = end_time - start_time

    # plot convergence graph
    fig, axes = Conv_plot(convergence_curve)
    if(save_conv_graph):
        plt.savefig('convergence_graph_'+ short_name + '.jpg')
    plt.show()

    # update attributes of solution
    solution.best_agent = Leader_agent
    solution.best_fitness = Leader_fitness
    solution.best_accuracy = Leader_accuracy
    solution.convergence_curve = convergence_curve
    solution.final_particles = particles
    solution.final_fitness = fitness
    solution.final_accuracy = accuracy
    solution.execution_time = exec_time

    return solution


if __name__ == '__main__':
    
    data = datasets.load_digits()
    PSO(20, 100, data.data, data.target, save_conv_graph=True)




from wrapper.nature_inspired.BBA import BBA
from wrapper.nature_inspired.CS import CS
from wrapper.nature_inspired.EO import EO
from wrapper.nature_inspired.GA import GA
from wrapper.nature_inspired.GSA import GSA
from wrapper.nature_inspired.GWO import GWO
from wrapper.nature_inspired.HS import HS
from wrapper.nature_inspired.MA import MA
from wrapper.nature_inspired.PSO import PSO
from wrapper.nature_inspired.RDA import RDA
from wrapper.nature_inspired.SCA import SCA
from wrapper.nature_inspired.WOA import WOA

__all__ = [
    "BBA",
    "CS",
    "EO",
    "GA",
    "GSA",
    "GWO",
    "HS",
    "MA",
    "PSO",
    "RDA",
    "SCA",
    "WOA"
]


import os,sys

#  set the directory path
dir_path = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, dir_path)

import wrapper
import filter
import evaluation
import datasets

__all__ = [
    "wrapper",
    "filter",
    "evaluation",
]



"""

Programmer: Ritam Guha
Date of Development: 28/10/2020

"""
import numpy as np
from filter._utilities import normalize, Result
from sklearn import datasets

def PCC(data, target):
    # function that assigns scores to features according to Pearson's Correlation Coefficient (PCC)
    # the rankings should be done in increasing order of the PCC scores 
    
    # initialize the variables and result structure
    feature_values = np.array(data)
    num_features = feature_values.shape[1]
    PCC_mat = np.zeros((num_features, num_features))
    PCC_values_feat = np.zeros(num_features)
    PCC_values_class = np.zeros(num_features)
    PCC_scores = np.zeros(num_features)
    result = Result()
    result.features = feature_values
    weight_feat = 0.3   # weightage provided to feature-feature correlation
    weight_class = 0.7  # weightage provided to feature-class correlation

    # generate the correlation matrix
    mean_values = np.mean(feature_values, axis=0)
    for ind_1 in range(num_features):
        for ind_2 in range(num_features):
            PCC_mat[ind_1, ind_2] = PCC_mat[ind_2, ind_1] = compute_PCC(feature_values[:, ind_1], feature_values[:, ind_2])

    for ind in range(num_features):
        PCC_values_feat[ind] = -np.sum(abs(PCC_mat[ind,:]))
        PCC_values_class[ind] = abs(compute_PCC(feature_values[:, ind], target))

    # produce scores and ranks from the information matrix
    PCC_values_feat = normalize(PCC_values_feat)
    PCC_values_class = normalize(PCC_values_class)
    PCC_scores = (weight_class * PCC_values_class) + (weight_feat * PCC_values_feat)
    PCC_ranks = np.argsort(np.argsort(-PCC_scores))

    # assign the results to the appropriate fields
    result.scores = PCC_scores
    result.ranks = PCC_ranks
    result.ranked_features = feature_values[:, np.argsort(-PCC_scores)]

    return result


def compute_PCC(x, y):
    # function to compute the PCC value for two variables
    mean_x = np.mean(x)
    mean_y = np.mean(y)
    numerator = np.sum((x - mean_x) * (y - mean_y))
    denominator = np.sqrt(np.sum(np.square(x - mean_x)) * np.sum(np.square(y - mean_y)))
    if denominator == 0:
        return 0
    PCC_val = numerator/denominator

    return PCC_val


if __name__ == '__main__':
    data = datasets.load_digits()
    PCC(data.data, data.target)


import numpy as np
from sklearn.neighbors import KNeighborsClassifier as KNN
import matplotlib.pyplot as plt

from sklearn.model_selection import train_test_split
from sklearn import datasets


class Solution():    
    #structure of the solution 
    def __init__(self):
        self.num_features = None
        self.num_agents = None
        self.max_iter = None
        self.obj_function = None
        self.execution_time = None
        self.best_agent = None
        self.best_fitness = None
        self.best_accuracy = None
        self.final_population = None
        self.final_fitness = None
        self.final_accuracy = None
        self.history = None


class Data():
    # structure of the training data
    def __init__(self):
        self.train_X = None
        self.train_Y = None
        self.val_X = None
        self.val_Y = None



def initialize(num_agents, num_features):  
    # create the population
  
    # define min and max number of features
    min_features = int(0.3 * num_features)
    max_features = int(0.6 * num_features)

    # initialize the agents with zeros
    agents = np.zeros((num_agents, num_features))

    # select random features for each agent
    for agent_no in range(num_agents):

        # find random indices
        cur_count = np.random.randint(min_features, max_features)
        temp_vec = np.random.rand(1, num_features)
        temp_idx = np.argsort(temp_vec)[0][0:cur_count]

        # select the features with the ranom indices
        agents[agent_no][temp_idx] = 1   

    return agents



def sort_agents(agents, fitness):
    # sort the agents according to fitness
    idx = np.argsort(-fitness)
    sorted_agents = agents[idx].copy()
    sorted_fitness = fitness[idx].copy()

    return sorted_agents, sorted_fitness



def display(agents, fitness, agent_name='Agent'):
    # display the population
    print('\nNumber of agents: {}'.format(agents.shape[0]))
    print('\n------------- Best Agent ---------------')
    print('Fitness: {}'.format(fitness[0]))
    print('Number of Features: {}'.format(int(np.sum(agents[0]))))
    print('----------------------------------------\n')

    for id, agent in enumerate(agents):
        print('{} {} - Fitness: {}, Number of Features: {}'.format(agent_name, id+1, fitness[id], int(np.sum(agent))))

    print('================================================================================\n')



def compute_accuracy(agents, data): 
    # compute classification accuracy of the given agents
    if len(agents.shape) == 1:
        # handles with zero-rank arrays
        agents = np.array([agents])

    (num_agents, num_features) = agents.shape
    acc = np.zeros(num_agents)   
    clf = KNN()
    
    for (i, agent) in enumerate(agents):
        cols = np.flatnonzero(agent) 

        if cols.size > 0:    
            train_data = data.train_X[:,cols]
            train_label = data.train_Y
            test_data = data.val_X[:,cols]
            test_label = data.val_Y

            clf.fit(train_data,train_label)
            acc[i] = clf.score(test_data,test_label)

    return acc
        

def compute_fitness(weight_acc):
    def _compute_fitness(agents, data):
        # compute a basic fitness measure
        
        if len(agents.shape) == 1:
        # handles with zero-rank arrays
            agents = np.array([agents])

        weight_feat = 1 - weight_acc
        (num_agents, num_features) = agents.shape
        fitness = np.zeros(num_agents)

        acc = compute_accuracy(agents, data)

        for (i, agent) in enumerate(agents):
            if np.sum(agents[i])!=0:
                feat = (num_features - np.sum(agents[i]))/num_features
                fitness[i] = weight_acc * acc[i] + weight_feat * feat

        return fitness
    
    return _compute_fitness



def call_counter(func):
    # meta function to count the number of calls to another function
	def helper(*args, **kwargs):
		helper.cur_evals += 1	
		func_val = func(*args, **kwargs)
		return func_val

	helper.cur_evals = 0
	helper.__name__= func.__name__
	return helper

